#include "matching.h"

#define ABS(x)    (((x) > 0) ? (x) : (-(x)))

using namespace std;

void compute_local_descriptor_keypoints(vector<float>& queryImg, int w, int h, int c, Keypointlist& KPs, int ps)
{
    if(!queryImg.empty())
    {
        keypointslist* keys = new keypointslist;
        siftPar siftparameters;
        default_sift_parameters(siftparameters);
        NewOriSize1 = ps;

        compute_sift_keypoints(queryImg.data(), *keys, w, h, c, siftparameters);
        KPs.resize(keys->size());
        for(int i=0; i<keys->size();i++)
        {
            KPs[i].x = (*keys)[i].x;
            KPs[i].y = (*keys)[i].y;
            KPs[i].kp_ptr = &((*keys)[i]);
            KPs[i].scale = (*keys)[i].scale;
            KPs[i].angle = (*keys)[i].angle;
        }
    }
}

/**
 * @brief Each SIIM (Scale Invariant Image Matching) descriptor in <keys> is added to its corresponding generalised descriptor. Already created generalised descriptors are stored in <mapKP>.
 * If a SIIM keypoint doen't correponds to a generalised descriptor in <mapKP> then it is created.
 * @param keys List of SIIM descriptors to be added
 * @param mapKP A map to already created generalised descriptors. For each pixel in the image there is possible a generalised descriptor,
 * so the size of the image is equal to the size of the map.
 * @param width mapKP width
 * @param height mapKP height
 * @author Mariano Rodríguez
 */
void add_keypoint(Keypointlist& keys, std::vector<KeyPoints*>& mapKP, int width, int height)
{
    float x,y;
    int xr,yr;
    bool only_center;
    for(int i=0; i<(int)keys.size();i++)
    {
        x = keys[i].x;
        xr = (int) round(x);
        y = keys[i].y;
        yr = (int) round(y);

        int ind =  yr*width + xr, newind;
        newind = ind;
        if ( mapKP[ind]==0 )
        {
            mapKP[ind] = new KeyPoints();
            mapKP[ind]->x = x;
            mapKP[ind]->y = y;
            mapKP[ind]->sum_x = x;
            mapKP[ind]->sum_y = y;
            mapKP[ind]->KPvec.push_back(keys[i]);
        }
        else
        {
            mapKP[ind]->KPvec.push_back(keys[i]);
            mapKP[ind]->sum_x += x;
            mapKP[ind]->sum_y += y;
            mapKP[ind]->x = mapKP[ind]->sum_x / mapKP[ind]->KPvec.size();
            mapKP[ind]->y = mapKP[ind]->sum_y / mapKP[ind]->KPvec.size();

        }

        only_center = false;
        int r = FILTER_RADIUS;
        while(!only_center)
        {
            only_center = true;
            for (int xi = xr-r; xi<=xr+r; xi++)
                for (int yi = yr-r; yi<=yr+r; yi++)
                {
                    if (( sqrt(pow(xi-xr,2) + pow(yi-yr,2))<=r )&&(xi>0)&&(xi<width)&&(yi>0)&&(yi<height))
                    {
                        int indi = yi*width + xi;
                        if ( (mapKP[indi]!=0)&&(indi!=ind) )
                        {
                            //merge indi to ind
                            only_center = false;
                            mapKP[ind]->sum_x += mapKP[indi]->sum_x;
                            mapKP[ind]->sum_y += mapKP[indi]->sum_y;

                            for (int k=0;k<(int)mapKP[indi]->KPvec.size();k++)
                            {
                                mapKP[ind]->KPvec.push_back(mapKP[indi]->KPvec[k]);
                            }
                            mapKP[indi] = 0;

                            mapKP[ind]->x = mapKP[ind]->sum_x / mapKP[ind]->KPvec.size();
                            mapKP[ind]->y = mapKP[ind]->sum_y / mapKP[ind]->KPvec.size();

                            //update newind
                            x = mapKP[ind]->x;
                            xr = (int) round(x);
                            y = mapKP[ind]->y;
                            yr = (int) round(y);
                            newind = yr*width + xr;

                            if (newind!=ind)
                            {
                                if (mapKP[newind]==0)
                                {// newind empty
                                    mapKP[newind] = mapKP[ind];
                                    mapKP[ind] = 0;
                                    ind = newind;
                                }
                                else
                                { // newind not empty
                                    mapKP[newind]->sum_x += mapKP[ind]->sum_x;
                                    mapKP[newind]->sum_y += mapKP[ind]->sum_y;

                                    for (int k=0;k<(int)mapKP[ind]->KPvec.size();k++)
                                    {
                                        mapKP[newind]->KPvec.push_back(mapKP[ind]->KPvec[k]);
                                    }

                                    mapKP[newind]->x = mapKP[newind]->sum_x / mapKP[newind]->KPvec.size();
                                    mapKP[newind]->y = mapKP[newind]->sum_y / mapKP[newind]->KPvec.size();
                                    mapKP[ind] = 0;
                                    ind = newind;
                                }
                            }
                        }
                    }

                }
        }

    }

}


/**
 * @brief Computes all generalised keypoints comming from a set of optical tilts digitally generated.
 * @param image Input image.
 * @param width Width of the input image.
 * @param height Height of the input image.
 * @param imasKP Returns a list of generalised keypoints.
 * @param simu_details Specifies the optical tilts that are to be simulated.
 * @param stats A vector with statistics on found generalised keypoints. Mean, min or max of SIIM keypoints over all found generalised keypoints.
 * @return The total number of generalised keypoints that have been found.
 * @author Mariano Rodríguez
 */
int compute_keypoints(vector<float>& image, int width, int height, int channels, std::vector<KeyPoints*>& keypoints,std::vector<float>& stats, int ps)
{	

    std::vector<KeyPoints*> mapKP;
    mapKP.resize(width*height);

    for(int i=0;i<width*height;i++)
        mapKP[0]= 0;

    int num_keys_total=0;

    Keypointlist keys;
    compute_local_descriptor_keypoints(image, width, height, channels, keys, ps);
    add_keypoint(keys, mapKP, width, height);

    // save in `keypoints` and do stats
    int num_max = 0, num_min = 500000, total = 0;
    float num_mean = 0;
    for (int i = 0; i < (int) mapKP.size(); i++)
        if (mapKP[i]!=0)
        {
            keypoints.push_back(mapKP[i]);
            total +=mapKP[i]->KPvec.size();
            num_keys_total += 1;
            if (num_max<(int)mapKP[i]->KPvec.size())
                num_max = mapKP[i]->KPvec.size();
            if (num_min>(int)mapKP[i]->KPvec.size())
                num_min = mapKP[i]->KPvec.size();
            num_mean +=mapKP[i]->KPvec.size();
        }
    num_mean = num_mean/num_keys_total;
    stats.push_back((float)total);
    stats.push_back((float)num_min);
    stats.push_back(num_mean);
    stats.push_back((float)num_max);

    return num_keys_total;
}


/**
 * @brief Computes matches among generalised descriptors coming from query and target images as described in \cite imas_IPOL_2017
 * @param w Width of image1
 * @param h Height of image1
 * @param keys Keypoints and generalised descriptors found on all simulated optical tilts of query image
 * @param matchings Returns a vector of matches after filtering
 * @return Total number of matches
 * @author Mariano Rodríguez
 */
int compute_matches(int c, std::vector<KeyPoints*>& keys, Matchingslist &matchings, int ps, float epsilon, bool automatic)
{  	
    int tstart = time(0);
    printf("Keypoints matching...\n");

    double tau;
    if(automatic)
        tau = 2*stats::qchisq(pow(exp(epsilon)/(keys.size()*keys.size()), 1./(ps*ps*c)), 1);
    else
        tau = epsilon;

    // Test every pair of descriptors
    for (int idx1 = 0; idx1 < keys.size(); ++idx1)
    for (int idx2 = 0; idx2 < idx1; ++idx2)
    {
        // Avoid matching a descriptor with itself
        if(idx1 == idx2)
            continue;

        bool stopped = false;
        for(int i1 = 0; i1 < keys[idx1]->KPvec.size() && !stopped; ++i1)
        for(int i2 = 0; i2 < keys[idx2]->KPvec.size(); ++i2)
        {
            // Try to match both descriptors and when one if flipped 
            if(patch_comparison(
                        static_cast<keypoint*>(keys[idx1]->KPvec[i1].kp_ptr)->gradx,
                        static_cast<keypoint*>(keys[idx1]->KPvec[i1].kp_ptr)->grady,
                        static_cast<keypoint*>(keys[idx2]->KPvec[i2].kp_ptr)->gradx,
                        static_cast<keypoint*>(keys[idx2]->KPvec[i2].kp_ptr)->grady,
                        ps, c, tau, false) ||
                    patch_comparison(
                        static_cast<keypoint*>(keys[idx1]->KPvec[i1].kp_ptr)->gradx,
                        static_cast<keypoint*>(keys[idx1]->KPvec[i1].kp_ptr)->grady,
                        static_cast<keypoint*>(keys[idx2]->KPvec[i2].kp_ptr)->gradx,
                        static_cast<keypoint*>(keys[idx2]->KPvec[i2].kp_ptr)->grady,
                        ps, c, tau, true))
            {
                Keypoint_simple k1, k2;

                k1.x = keys[idx1]->x;
                k1.y = keys[idx1]->y;
                k1.scale = keys[idx1]->KPvec[i1].scale;
                k1.angle = keys[idx1]->KPvec[i1].angle;

                k2.x = keys[idx2]->x;
                k2.y = keys[idx2]->y;
                k2.scale = keys[idx2]->KPvec[i2].scale;
                k2.angle = keys[idx2]->KPvec[i2].angle;

                matchings.push_back( Matching(k1,k2) );
                stopped = true;
                break;
            }
        }
    }

    printf("   %d matches found.\n", (int) matchings.size());
    printf("Matching accomplished in %ld seconds.\n\n", (time(0) - tstart));
    return matchings.size();
}

void perform_matching(int channels, vector<float>& image, int w, int h, vector<float>& data, Matchingslist& matchings, int ps, float tau, bool automatic)
{
    // Compute keypoints
    std::vector<KeyPoints*> keys;
    keys.clear();

    int num_keys=0;

    int tstart = time(0);
    std::vector<float> stats;
    num_keys = compute_keypoints(image, w, h, channels, keys, stats, ps);
    printf("%d keypoints have been found\n", num_keys);
    printf("stats: group_min = %d , group_mean = %.3f, group_max = %d\n",(int)stats[1],stats[2],(int)stats[3]);
    printf("Keypoints computation accomplished in %ld seconds.\n \n", time(0) - tstart);


    // Match keypoints
    compute_matches(channels, keys, matchings, ps, tau, automatic);

    /* Generate data matrix: matchinglist and Minfoall */
    for ( int i = 0; i < (int) matchings.size(); i++ )
    {
        Matching *ptr_in = &(matchings[i]);
        // there are 8 rows of info
        data.push_back(ptr_in->first.x);
        data.push_back(ptr_in->first.y);
        data.push_back(ptr_in->first.scale);
        data.push_back(ptr_in->first.angle);

        data.push_back(ptr_in->second.x);
        data.push_back(ptr_in->second.y);
        data.push_back(ptr_in->second.scale);
        data.push_back(ptr_in->second.angle);
    }


    printf("Done.\n\n");
    keys.clear();
}
